from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from loopy.auto_test import auto_test_vs_ref
from loopy.codegen import PreambleInfo, generate_body, generate_code, generate_code_v2
from loopy.codegen.result import CodeGenerationResult, GeneratedProgram
from loopy.diagnostic import LoopyError, LoopyWarning
from loopy.frontend.fortran import (
    c_preprocess,
    parse_fortran,
    parse_transformed_fortran,
)
from loopy.kernel import KernelState, LoopKernel
from loopy.kernel.creation import UniqueName, make_function, make_kernel
from loopy.kernel.data import (
    AddressSpace,
    ArrayArg,
    CallMangleInfo,
    ConstantArg,
    GlobalArg,
    ImageArg,
    KernelArgument,
    SubstitutionRule,
    TemporaryVariable,
    ValueArg,
)
from loopy.kernel.function_interface import (
    CallableKernel,
    InKernelCallable,
    ScalarCallable,
)
from loopy.kernel.instruction import (
    Assignment,
    AtomicInit,
    AtomicUpdate,
    BarrierInstruction,
    CallInstruction,
    CInstruction,
    HappensAfter,
    InstructionBase,
    LegacyStringInstructionTag,
    MemoryOrdering,
    MemoryScope,
    MultiAssignmentBase,
    NoOpInstruction,
    OrderedAtomic,
    UseStreamingStoreTag,
    VarAtomicity,
)
from loopy.kernel.tools import (
    add_and_infer_dtypes,
    add_dtypes,
    find_most_recent_global_barrier,
    get_dot_dependency_graph,
    get_global_barrier_order,
    get_subkernel_to_insn_id_map,
    get_subkernels,
    show_dependency_graph,
)
from loopy.library.reduction import register_reduction_parser
from loopy.options import Options
from loopy.preprocess import infer_arg_descr, preprocess_kernel, preprocess_program
from loopy.schedule import (
    generate_loop_schedules,
    get_one_linearized_kernel,
    get_one_scheduled_kernel,
    linearize,
)
from loopy.statistics import (
    CountGranularity,
    MemAccess,
    Op,
    Sync,
    ToCountMap,
    ToCountPolynomialMap,
    gather_access_footprint_bytes,
    gather_access_footprints,
    get_mem_access_map,
    get_op_map,
    get_synchronization_map,
)
from loopy.symbolic import LinearSubscript, Reduction, TaggedVariable, TypeCast
from loopy.target import ASTBuilderBase, TargetBase
from loopy.target.c import (
    CFamilyTarget,
    CTarget,
    CWithGNULibcTarget,
    ExecutableCTarget,
    ExecutableCWithGNULibcTarget,
    generate_header,
)
from loopy.target.cuda import CudaTarget
from loopy.target.execution import ExecutorBase
from loopy.target.ispc import ISPCTarget
from loopy.target.opencl import OpenCLTarget
from loopy.target.pyopencl import PyOpenCLTarget
from loopy.tools import Optional, clear_in_mem_caches, memoize_on_disk, t_unit_to_python
from loopy.transform.add_barrier import add_barrier
from loopy.transform.arithmetic import (
    collect_common_factors_on_increment,
    fold_constants,
)
from loopy.transform.batch import to_batched
from loopy.transform.buffer import buffer_array
from loopy.transform.callable import (
    inline_callable_kernel,
    merge,
    register_callable,
    rename_callable,
)
from loopy.transform.concatenate import concatenate_arrays
from loopy.transform.data import (
    add_prefetch,
    alias_temporaries,
    allocate_temporaries_for_base_storage,
    change_arg_to_image,
    remove_unused_arguments,
    rename_argument,
    set_argument_order,
    set_array_axis_names,
    set_array_dim_names,
    set_temporary_address_space,
    set_temporary_scope,
    tag_array_axes,
    tag_data_axes,
)
from loopy.transform.fusion import fuse_kernels
from loopy.transform.iname import (
    add_inames_for_unused_hw_axes,
    add_inames_to_insn,
    affine_map_inames,
    chunk_iname,
    duplicate_inames,
    find_unused_axis_tag,
    get_iname_duplication_options,
    has_schedulable_iname_nesting,
    join_inames,
    make_reduction_inames_unique,
    map_domain,
    prioritize_loops,
    remove_inames_from_insn,
    remove_predicates_from_insn,
    remove_unused_inames,
    rename_iname,
    rename_inames,
    split_iname,
    split_reduction_inward,
    split_reduction_outward,
    tag_inames,
    untag_inames,
)
from loopy.transform.instruction import (
    add_dependency,
    add_nosync,
    find_instructions,
    map_instructions,
    remove_instructions,
    replace_instruction_ids,
    set_instruction_priority,
    simplify_indices,
    tag_instructions,
)
from loopy.transform.pack_and_unpack_args import pack_and_unpack_args_for_call
from loopy.transform.padding import (
    add_padding,
    find_padding_multiple,
    split_arg_axis,
    split_array_axis,
    split_array_dim,
)
from loopy.transform.parameter import assume, fix_parameters
from loopy.transform.precompute import precompute
from loopy.transform.privatize import (
    privatize_temporaries_with_inames,
    unprivatize_temporaries_with_inames,
)
from loopy.transform.realize_reduction import realize_reduction
from loopy.transform.save import save_and_reload_temporaries
from loopy.transform.subst import (
    assignment_to_subst,
    expand_subst,
    extract_subst,
    find_one_rule_matching,
    find_rules_matching,
)
from loopy.translation_unit import TranslationUnit, for_each_kernel, make_program
from loopy.type_inference import infer_unknown_types
from loopy.types import LoopyType, NumpyType, ToLoopyTypeConvertible, to_loopy_type
from loopy.typing import auto
from loopy.version import MOST_RECENT_LANGUAGE_VERSION, VERSION


__all__ = [
    "MOST_RECENT_LANGUAGE_VERSION",
    "VERSION",
    "ASTBuilderBase",
    "AddressSpace",
    "ArrayArg",
    "Assignment",
    "AtomicInit",
    "AtomicUpdate",
    "BarrierInstruction",
    "CFamilyTarget",
    "CInstruction",
    "CTarget",
    "CWithGNULibcTarget",
    "CacheMode",
    "CallInstruction",
    "CallMangleInfo",
    "CallableKernel",
    "CodeGenerationResult",
    "ConstantArg",
    "CountGranularity",
    "CudaTarget",
    "ExecutableCTarget",
    "ExecutableCWithGNULibcTarget",
    "ExecutorBase",
    "GeneratedProgram",
    "GlobalArg",
    "HappensAfter",
    "ISPCTarget",
    "ImageArg",
    "InKernelCallable",
    "InstructionBase",
    "KernelArgument",
    "KernelState",
    "LegacyStringInstructionTag",
    "LinearSubscript",
    "LoopKernel",
    "LoopyError",
    "LoopyType",
    "LoopyWarning",
    "MemAccess",
    "MemoryOrdering",
    "MemoryScope",
    "MultiAssignmentBase",
    "NoOpInstruction",
    "NumpyType",
    "Op",
    "OpenCLTarget",
    "Optional",
    "Options",
    "OrderedAtomic",
    "PreambleInfo",
    "PyOpenCLTarget",
    "Reduction",
    "ScalarCallable",
    "SubstitutionRule",
    "Sync",
    "TaggedVariable",
    "TargetBase",
    "TemporaryVariable",
    "ToCountMap",
    "ToCountPolynomialMap",
    "ToLoopyTypeConvertible",
    "TranslationUnit",
    "TypeCast",
    "UniqueName",
    "UseStreamingStoreTag",
    "ValueArg",
    "VarAtomicity",
    "add_and_infer_dtypes",
    "add_barrier",
    "add_dependency",
    "add_dtypes",
    "add_inames_for_unused_hw_axes",
    "add_inames_to_insn",
    "add_nosync",
    "add_padding",
    "add_prefetch",
    "affine_map_inames",
    "alias_temporaries",
    "allocate_temporaries_for_base_storage",
    "assignment_to_subst",
    "assume",
    "auto",
    "auto_test_vs_ref",
    "buffer_array",
    "c_preprocess",
    "change_arg_to_image",
    "chunk_iname",
    "clear_in_mem_caches",
    "collect_common_factors_on_increment",
    "concatenate_arrays",
    "duplicate_inames",
    "expand_subst",
    "extract_subst",
    "find_instructions",
    "find_most_recent_global_barrier",
    "find_one_rule_matching",
    "find_padding_multiple",
    "find_rules_matching",
    "find_unused_axis_tag",
    "fix_parameters",
    "fold_constants",
    "for_each_kernel",
    "fuse_kernels",
    "gather_access_footprint_bytes",
    "gather_access_footprints",
    "generate_body",
    "generate_code",
    "generate_code_v2",
    "generate_header",
    "generate_loop_schedules",
    "get_dot_dependency_graph",
    "get_global_barrier_order",
    "get_iname_duplication_options",
    "get_mem_access_map",
    "get_one_linearized_kernel",
    "get_one_scheduled_kernel",
    "get_op_map",
    "get_subkernel_to_insn_id_map",
    "get_subkernels",
    "get_synchronization_map",
    "has_schedulable_iname_nesting",
    "infer_arg_descr",
    "infer_unknown_types",
    "inline_callable_kernel",
    "join_inames",
    "linearize",
    "make_copy_kernel",
    "make_einsum",
    "make_function",
    "make_kernel",
    "make_kernel",
    "make_program",
    "make_reduction_inames_unique",
    "map_domain",
    "map_instructions",
    "memoize_on_disk",
    "merge",
    "pack_and_unpack_args_for_call",
    "parse_fortran",
    "parse_transformed_fortran",
    "precompute",
    "preprocess_kernel",
    "preprocess_program",
    "prioritize_loops",
    "privatize_temporaries_with_inames",
    "realize_reduction",
    "register_callable",
    "register_preamble_generators",
    "register_reduction_parser",
    "register_symbol_manglers",
    "remove_inames_from_insn",
    "remove_instructions",
    "remove_predicates_from_insn",
    "remove_unused_arguments",
    "remove_unused_inames",
    "rename_argument",
    "rename_callable",
    "rename_iname",
    "rename_inames",
    "replace_instruction_ids",
    "save_and_reload_temporaries",
    "set_argument_order",
    "set_array_axis_names",
    "set_array_dim_names",
    "set_caching_enabled",
    "set_instruction_priority",
    "set_temporary_address_space",
    "set_temporary_scope",
    "show_dependency_graph",
    "simplify_indices",
    "split_arg_axis",
    "split_array_axis",
    "split_array_dim",
    "split_iname",
    "split_reduction_inward",
    "split_reduction_outward",
    "t_unit_to_python",
    "tag_array_axes",
    "tag_data_axes",
    "tag_inames",
    "tag_instructions",
    "to_batched",
    "to_loopy_type",
    "unprivatize_temporaries_with_inames",
    "untag_inames",
    ]

# }}}


# {{{ set_options

@for_each_kernel
def set_options(kernel, *args, **kwargs):
    """Return a new kernel with the options given as keyword arguments, or from
    a string representation passed in as the first (and only) positional
    argument.

    See also :class:`Options`.
    """
    assert isinstance(kernel, LoopKernel)

    if args and kwargs:
        raise TypeError("cannot pass both positional and keyword arguments")

    new_opt = kernel.options.copy()

    if kwargs:
        from loopy.options import Options, _apply_legacy_map
        kwargs = _apply_legacy_map(Options._legacy_options_map, kwargs)

        for key, val in kwargs.items():
            if not hasattr(new_opt, key):
                raise ValueError("unknown option '%s'" % key)

            setattr(new_opt, key, val)
    else:
        if len(args) != 1:
            raise TypeError("exactly one positional argument is required if "
                    "no keyword args are given")
        arg, = args

        from loopy.options import make_options
        new_opt._update(make_options(arg))

    return kernel.copy(options=new_opt)

# }}}


# {{{ library registration

@for_each_kernel
def register_preamble_generators(kernel: LoopKernel, preamble_generators):
    """
    :arg manglers: list of functions of signature ``(preamble_info)``
        generating tuples ``(sortable_str_identifier, code)``,
        where *preamble_info* is a :class:`PreambleInfo`.

    :returns: *kernel* with *manglers* registered
    """
    from loopy.tools import unpickles_equally

    new_pgens = tuple(kernel.preamble_generators)

    for pgen in preamble_generators:
        if pgen not in new_pgens:
            if not unpickles_equally(pgen):
                raise LoopyError("preamble generator '%s' does not "
                        "compare equally after being unpickled "
                        "and would thus disrupt loopy's caches"
                        % pgen)

            new_pgens = (pgen, *new_pgens)

    return kernel.copy(preamble_generators=new_pgens)


@for_each_kernel
def register_symbol_manglers(kernel, manglers):
    from loopy.tools import unpickles_equally

    new_manglers = kernel.symbol_manglers
    for m in manglers:
        if m not in new_manglers:
            if not unpickles_equally(m):
                raise LoopyError("mangler '%s' does not "
                        "compare equally after being unpickled "
                        "and would disrupt loopy's caches"
                        % m)

            new_manglers = (m, *new_manglers)

    return kernel.copy(symbol_manglers=new_manglers)

# }}}


# {{{ cache control

import os

from pytools import strtobool


# Caching is enabled by default, but can be disabled by setting
# the environment variables LOOPY_NO_CACHE or CG_NO_CACHE to a
# 'true' value.
CACHING_ENABLED = (
    not strtobool(os.environ.get("LOOPY_NO_CACHE", "false"))
    and
    not strtobool(os.environ.get("CG_NO_CACHE", "false")))


ABORT_ON_CACHE_MISS = strtobool(os.environ.get("LOOPY_ABORT_ON_CACHE_MISS", "false"))


def set_caching_enabled(flag):
    """Set whether :mod:`loopy` is allowed to use disk caching for its various
    code generation stages.
    """
    global CACHING_ENABLED
    CACHING_ENABLED = flag


class CacheMode:
    """A context manager for setting whether :mod:`loopy` is allowed to use
    disk caches.
    """

    def __init__(self, new_flag):
        self.new_flag = new_flag

    def __enter__(self):
        global CACHING_ENABLED
        self.previous_mode = CACHING_ENABLED
        CACHING_ENABLED = self.new_flag

    def __exit__(self, exc_type, exc_val, exc_tb):
        global CACHING_ENABLED
        CACHING_ENABLED = self.previous_mode
        del self.previous_mode

# }}}


# {{{ make copy kernel

def make_copy_kernel(new_dim_tags, old_dim_tags=None):
    """Returns a :class:`loopy.TranslationUnit` that changes the data layout
    of a variable (called "input") to the new layout specified by
    *new_dim_tags* from the one specified by *old_dim_tags*.
    *old_dim_tags* defaults to an all-C layout of the same rank
    as the one given by *new_dim_tags*.
    """

    from loopy.kernel.array import (
        SeparateArrayArrayDimTag,
        VectorArrayDimTag,
        parse_array_dim_tags,
    )
    new_dim_tags = parse_array_dim_tags(new_dim_tags, n_axes=None)

    rank = len(new_dim_tags)
    if old_dim_tags is None:
        old_dim_tags = parse_array_dim_tags(
                ",".join(rank * ["c"]), n_axes=None)
    elif isinstance(old_dim_tags, str):
        old_dim_tags = parse_array_dim_tags(
                old_dim_tags, n_axes=None)

    indices = ["i%d" % i for i in range(rank)]
    shape = ["n%d" % i for i in range(rank)]
    command_indices = ", ".join(indices)
    bounds = " and ".join(
            f"0<={ind}<{shape_i}"
            for ind, shape_i in zip(indices, shape))

    set_str = "{{[{}]: {} }}".format(
                command_indices,
                bounds
                )
    result = make_kernel(set_str,
            "output[%s] = input[%s]"
            % (command_indices, command_indices),
            lang_version=MOST_RECENT_LANGUAGE_VERSION,
            default_offset=auto)

    result = tag_array_axes(result, "input", old_dim_tags)
    result = tag_array_axes(result, "output", new_dim_tags)

    unrolled_tags = (SeparateArrayArrayDimTag, VectorArrayDimTag)
    for i in range(rank):
        if (isinstance(new_dim_tags[i], unrolled_tags)
                or isinstance(old_dim_tags[i], unrolled_tags)):
            result = tag_inames(result, {indices[i]: "unr"})

    return result

# }}}


# {{{ einsum

def make_einsum(spec, arg_names, **knl_creation_kwargs):
    r"""Returns a :class:`LoopKernel` for evaluating array-based
    operations using Einstein summation convention.

    :param spec: a string denoting the subscripts for
        summation as a comma-separated list of subscript labels.
        This follows the usual :func:`numpy.einsum` convention.
        Note that the explicit indicator `->` for the precise output
        form is required.
    :param arg_names: a sequence of string types denoting
        the names of the array operands.
    :param \**knl_creation_kwargs: keyword arguments for kernel creation.
        See :func:`make_kernel` for a list of acceptable keyword
        parameters.

    .. note::

        No attempt is being made to reduce the complexity
        of the resulting expression. This should be dealt with
        as part of a separate transformation.
    """
    arg_spec, out_spec = spec.split("->")
    arg_specs = arg_spec.split(",")

    out_spec = out_spec.strip()
    arg_specs = [arg_spec.strip() for arg_spec in arg_specs]

    if len(arg_names) != len(arg_specs):
        raise ValueError(
            f"Number of arg names ({arg_names}) should match the number "
            f"of arg specs: {arg_specs}. Length of arg names is {len(arg_names)}; "
            f"expecting {len(arg_specs)} arg names."
        )

    out_indices = set(out_spec)
    if len(out_indices) != len(out_spec):
        raise ValueError(
            f"Output subscripts '{out_spec}' does not contain all unique indices."
        )

    all_indices = {
        idx
        for argsp in arg_specs
        for idx in argsp} | out_indices

    sum_indices = all_indices - out_indices

    from pymbolic import var
    lhs = var("out")[tuple(var(i) for i in out_spec)]

    rhs = 1
    for arg_name, argsp in zip(arg_names, arg_specs):
        rhs = rhs * var(arg_name)[tuple(var(i) for i in argsp)]

    if sum_indices:
        rhs = Reduction("sum", tuple(var(idx) for idx in sorted(sum_indices)), rhs)

    constraints = " and ".join(
        "0 <= %s < N%s" % (idx, idx)
        for idx in all_indices
        )

    knl_creation_kwargs.setdefault("name", "einsum%dto%d_kernel" % (
            len(all_indices), len(out_indices)))
    knl_creation_kwargs.setdefault("lang_version", MOST_RECENT_LANGUAGE_VERSION)

    return make_kernel("{[%s]: %s}" % (",".join(sorted(all_indices)), constraints),
                       [Assignment(lhs, rhs)],
                       **knl_creation_kwargs)

# }}}


# {{{ default target

_DEFAULT_TARGET = None


def set_default_target(target):
    # deliberately undocumented for now
    global _DEFAULT_TARGET
    _DEFAULT_TARGET = target


def _set_up_default_target():
    try:
        import pyopencl  # noqa
    except ImportError:
        from loopy.target.opencl import OpenCLTarget
        target = OpenCLTarget
    else:
        from loopy.target.pyopencl import PyOpenCLTarget
        target = PyOpenCLTarget

    set_default_target(target)


_set_up_default_target()

# }}}


# vim: foldmethod=marker



import numpy as np
from constantdict import constantdict

import loopy as lp


class NoRetFunction(lp.ScalarCallable):
    def with_types(self, arg_id_to_dtype, callables):
        if len(arg_id_to_dtype) != 0:
            raise RuntimeError("'f' cannot take any inputs.")

        return (self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype),
                         name_in_target="f"),
                callables)

    def with_descrs(self, arg_id_to_descr, callables):
        if len(arg_id_to_descr) != 0:
            raise RuntimeError("'f' cannot take any inputs.")

        return (self.copy(arg_id_to_descr=constantdict(arg_id_to_descr)),
                callables)

    def generate_preambles(self, target):
        assert isinstance(target, lp.CFamilyTarget)
        yield ("10_define_f",
                r"""
                void f()
                {
                    printf("Hi!\n");
                }
                """)


class SingleArgNoRetFunction(lp.ScalarCallable):
    def with_types(self, arg_id_to_dtype, callables):
        input_dtype = arg_id_to_dtype.get(0)
        if input_dtype is None:
            return self, callables

        if input_dtype.numpy_dtype != np.float32:
            raise RuntimeError("'f' only supports f32.")

        return (self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype),
                          name_in_target="f"),
                callables)

    def with_descrs(self, arg_id_to_descr, callables):
        if len(arg_id_to_descr) != 0:
            raise RuntimeError("'f' cannot take any inputs.")

        return (self.copy(arg_id_to_descr=arg_id_to_descr),
                callables)

    def generate_preambles(self, target):
        assert isinstance(target, lp.CFamilyTarget)

        yield ("10_define_f",
                r"""
                void f(float x)
                {
                    printf("Hi!\n");
                }
                """)


def symbol_x(knl, name):
    if name == "X":
        from loopy.types import to_loopy_type
        return to_loopy_type(np.float32), "X"


def preamble_for_x(preamble_info):
    yield ("preamble_ten", r"#define X 10.0")



"""OpenCL target independent of PyOpenCL."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import TYPE_CHECKING, Literal, Sequence

import numpy as np
from constantdict import constantdict

from pymbolic import var
from pytools import memoize_method

from loopy.diagnostic import LoopyError, LoopyTypeError
from loopy.kernel.array import ArrayBase, FixedStrideArrayDimTag, VectorArrayDimTag
from loopy.kernel.data import AddressSpace, ConstantArg, ImageArg
from loopy.kernel.function_interface import ScalarCallable
from loopy.target.c import CFamilyASTBuilder, CFamilyTarget, DTypeRegistryWrapper
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper
from loopy.types import NumpyType


if TYPE_CHECKING:
    from cgen import Declarator, Generable

    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult


# {{{ dtype registry wrappers


class DTypeRegistryWrapperWithInt8ForBool(DTypeRegistryWrapper):
    """
    A DType registry that uses int8 for bool_ types.

    .. note::

        This sub-class is needed because compyte's type registry does
        not support type aliases.
    """
    def dtype_to_ctype(self, dtype):
        from loopy.types import NumpyType
        if isinstance(dtype, NumpyType) and dtype.dtype == np.bool_:
            return self.wrapped_registry.dtype_to_ctype(
                    NumpyType(np.int8))
        return self.wrapped_registry.dtype_to_ctype(dtype)


class DTypeRegistryWrapperWithAtomics(DTypeRegistryWrapper):
    def get_or_register_dtype(self, names, dtype=None):
        if dtype is not None:
            from loopy.types import AtomicNumpyType, NumpyType
            if isinstance(dtype, AtomicNumpyType):
                return self.wrapped_registry.get_or_register_dtype(
                        names, NumpyType(dtype.dtype))

        return self.wrapped_registry.get_or_register_dtype(names, dtype)


class DTypeRegistryWrapperWithCL1Atomics(DTypeRegistryWrapperWithAtomics):
    def dtype_to_ctype(self, dtype):
        from loopy.types import AtomicNumpyType

        if isinstance(dtype, AtomicNumpyType):
            return "volatile " + self.wrapped_registry.dtype_to_ctype(dtype)
        else:
            return self.wrapped_registry.dtype_to_ctype(dtype)

# }}}


# {{{ vector types

class vec:  # noqa
    pass


def _create_vector_types():
    field_names = ["x", "y", "z", "w"]

    vec.types = {}
    vec.names_and_dtypes = []
    vec.type_to_scalar_and_count = {}

    counts = [2, 3, 4, 8, 16]

    for base_name, base_type in [
            ("char", np.int8),
            ("uchar", np.uint8),
            ("short", np.int16),
            ("ushort", np.uint16),
            ("int", np.int32),
            ("uint", np.uint32),
            ("long", np.int64),
            ("ulong", np.uint64),
            ("float", np.float32),
            ("double", np.float64),
            ]:
        for count in counts:
            name = "%s%d" % (base_name, count)

            titles = field_names[:count]

            padded_count = count
            if count == 3:
                padded_count = 4

            names = ["s%d" % i for i in range(count)]
            while len(names) < padded_count:
                names.append("padding%d" % (len(names)-count))

            if len(titles) < len(names):
                titles.extend((len(names)-len(titles))*[None])

            try:
                dtype = np.dtype({
                    "names": names,
                    "formats": [base_type]*padded_count,
                    "titles": titles})
            except NotImplementedError:
                try:
                    dtype = np.dtype([((n, title), base_type)
                                      for (n, title) in zip(names, titles)])
                except TypeError:
                    dtype = np.dtype([(n, base_type) for (n, title)
                                      in zip(names, titles)])

            setattr(vec, name, dtype)

            vec.names_and_dtypes.append((name, dtype))

            vec.types[np.dtype(base_type), count] = dtype
            vec.type_to_scalar_and_count[dtype] = np.dtype(base_type), count


_create_vector_types()


def _register_vector_types(dtype_registry):
    for name, dtype in vec.names_and_dtypes:
        dtype_registry.get_or_register_dtype(name, dtype)

# }}}


# {{{ function mangler

_CL_SIMPLE_MULTI_ARG_FUNCTIONS = {
        "rsqrt": 1,
        "clamp": 3,
        "atan2": 2,
        }


VECTOR_LITERAL_FUNCS = {
        "make_%s%d" % (name, count): (name, dtype, count)
        for name, dtype in [
            ("char", np.int8),
            ("uchar", np.uint8),
            ("short", np.int16),
            ("ushort", np.uint16),
            ("int", np.int32),
            ("uint", np.uint32),
            ("long", np.int64),
            ("ulong", np.uint64),
            ("float", np.float32),
            ("double", np.float64),
            ]
        for count in [2, 3, 4, 8, 16]
        }


class OpenCLCallable(ScalarCallable):
    """
    Records information about OpenCL functions which are not covered by
    :class:`loopy.target.c.CMathCallable`.
    """

    def with_types(self, arg_id_to_dtype, callables_table):
        name = self.name

        # {{{ unary functions
        if name == "abs":
            for id in arg_id_to_dtype:
                if not -1 <= id <= 0:
                    raise LoopyError(f"'{name}' can take only one argument.")

            if 0 not in arg_id_to_dtype or arg_id_to_dtype[0] is None:
                return (
                        self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                        callables_table)

            dtype = arg_id_to_dtype[0].numpy_dtype

            if dtype.kind in ("u", "i"):
                # OpenCL C 2.2, Section 6.13.3: abs returns *u*gentype
                from loopy.types import to_unsigned_dtype
                return (self.copy(name_in_target=name,
                            arg_id_to_dtype=constantdict({
                                0: NumpyType(dtype),
                                -1: NumpyType(to_unsigned_dtype(dtype))
                                })),
                        callables_table)
            elif dtype.kind == "f":
                name = "fabs"
            else:
                raise LoopyTypeError(f"'{name}' does not support type {dtype}")

        # deliberately not elif: abs branch above may end up taking this.
        if name in ["fabs", "acos", "asin", "atan", "cos", "cosh", "sin", "sinh",
                    "tan", "tanh", "exp", "log", "log10", "sqrt", "ceil", "floor",
                    "erf", "erfc"]:

            for id in arg_id_to_dtype:
                if not -1 <= id <= 0:
                    raise LoopyError(f"'{name}' can take only one argument.")

            if 0 not in arg_id_to_dtype or arg_id_to_dtype[0] is None:
                return (
                        self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                        callables_table)

            dtype = arg_id_to_dtype[0]
            dtype = dtype.numpy_dtype

            if dtype.kind in ("u", "i"):
                # ints and unsigned casted to float32
                dtype = np.float32
            elif dtype.kind == "c":
                raise LoopyTypeError(f"{name} does not support type {dtype}")

            return (
                    self.copy(name_in_target=name,
                        arg_id_to_dtype=constantdict({
                            0: NumpyType(dtype),
                            -1: NumpyType(dtype)
                            })),
                    callables_table)

        # }}}

        # binary functions
        elif name in ["fmax", "fmin", "atan2", "copysign"]:

            for id in arg_id_to_dtype:
                if not -1 <= id <= 1:
                    # FIXME: Do we need to raise here?:
                    #   The pattern we generally follow is that if we don't find
                    #   a function, then we just return None
                    raise LoopyError("%s can take only two arguments." % name)

            if 0 not in arg_id_to_dtype or 1 not in arg_id_to_dtype or (
                    arg_id_to_dtype[0] is None or arg_id_to_dtype[1] is None):
                return (
                        self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                        callables_table)

            dtype = np.result_type(*[
                    dtype.numpy_dtype for id, dtype in arg_id_to_dtype.items()
                    if id >= 0])

            if dtype.kind == "c":
                raise LoopyTypeError(f"'{name}' does not support complex numbers")

            dtype = NumpyType(dtype)
            return (
                    self.copy(name_in_target=name,
                        arg_id_to_dtype=constantdict({
                            -1: dtype, 0: dtype, 1: dtype
                            })),
                    callables_table)

        elif name in ["max", "min"]:
            for id in arg_id_to_dtype:
                if not -1 <= id <= 1:
                    raise LoopyError("%s can take only 2 arguments." % name)
            if 0 not in arg_id_to_dtype or 1 not in arg_id_to_dtype:
                return (
                        self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                        callables_table)
            common_dtype = np.result_type(*[
                    dtype.numpy_dtype for id, dtype in arg_id_to_dtype.items()
                    if (id >= 0 and dtype is not None)])

            if common_dtype.kind in ["u", "i", "f"]:
                if common_dtype.kind == "f":
                    name = "f"+name

                dtype = NumpyType(common_dtype)
                return (
                        self.copy(name_in_target=name,
                            arg_id_to_dtype=constantdict({
                                -1: dtype, 0: dtype, 1: dtype
                                })),
                        callables_table)
            else:
                # Unsupported type.
                raise LoopyError("%s function not supported for the types %s" %
                        (name, common_dtype))

        elif name == "dot":
            for id in arg_id_to_dtype:
                if not -1 <= id <= 1:
                    raise LoopyError(f"'{name}' can take only 2 arguments.")

            if 0 not in arg_id_to_dtype or 1 not in arg_id_to_dtype or (
                    arg_id_to_dtype[0] is None or arg_id_to_dtype[1] is None):
                # the types provided aren't mature enough to specialize the
                # callable
                return (
                        self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                        callables_table)

            dtype = arg_id_to_dtype[0]
            scalar_dtype, _offset, _field_name = dtype.numpy_dtype.fields["s0"]
            return (
                    self.copy(name_in_target=name, arg_id_to_dtype=constantdict({
                        -1: NumpyType(scalar_dtype), 0: dtype, 1: dtype
                        })),
                    callables_table)

        elif name == "pow":
            for id in arg_id_to_dtype:
                if not -1 <= id <= 1:
                    raise LoopyError(f"'{name}' can take only 2 arguments.")

            common_dtype = np.result_type(*[
                    dtype.numpy_dtype for id, dtype in arg_id_to_dtype.items()
                    if (id >= 0 and dtype is not None)])

            if common_dtype == np.float64:
                name = "powf64"
            elif common_dtype == np.float32:
                name = "powf32"
            else:
                raise LoopyTypeError(f"'pow' does not support type {dtype}.")

            result_dtype = NumpyType(common_dtype)

            return (
                    self.copy(name_in_target=name,
                              arg_id_to_dtype=constantdict({
                                  -1: result_dtype,
                                  0: common_dtype,
                                  1: common_dtype
                                  })),
                    callables_table)

        elif name in _CL_SIMPLE_MULTI_ARG_FUNCTIONS:
            num_args = _CL_SIMPLE_MULTI_ARG_FUNCTIONS[name]
            for id in arg_id_to_dtype:
                if not -1 <= id < num_args:
                    raise LoopyError("%s can take only %d arguments." % (name,
                            num_args))

            for i in range(num_args):
                if i not in arg_id_to_dtype or arg_id_to_dtype[i] is None:
                    # the types provided aren't mature enough to specialize the
                    # callable
                    return (
                            self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                            callables_table)

            dtype = np.result_type(*[
                    dtype.numpy_dtype for id, dtype in arg_id_to_dtype.items()
                    if id >= 0])

            if dtype.kind == "c":
                raise LoopyError("%s does not support complex numbers"
                        % name)

            updated_arg_id_to_dtype = constantdict({
                id: NumpyType(dtype) for id in range(-1, num_args)
                })

            return (
                    self.copy(name_in_target=name,
                        arg_id_to_dtype=updated_arg_id_to_dtype),
                    callables_table)

        elif name in VECTOR_LITERAL_FUNCS:
            base_tp_name, dtype, count = VECTOR_LITERAL_FUNCS[name]

            for id in arg_id_to_dtype:
                if not -1 <= id < count:
                    raise LoopyError("%s can take only %d arguments." % (name,
                            num_args))

            for i in range(count):
                if i not in arg_id_to_dtype or arg_id_to_dtype[i] is None:
                    # the types provided aren't mature enough to specialize the
                    # callable
                    return (
                            self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                            callables_table)

            updated_arg_id_to_dtype = {id: NumpyType(dtype) for id in range(count)}
            updated_arg_id_to_dtype[-1] = OpenCLTarget().vector_dtype(
                        NumpyType(dtype), count)

            return (
                    self.copy(
                        name_in_target="(%s%d) " % (base_tp_name, count),
                        arg_id_to_dtype=constantdict(updated_arg_id_to_dtype)),
                    callables_table)

        # does not satisfy any of the conditions needed for specialization.
        # hence just returning a copy of the callable.
        return (
                self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                callables_table)


def get_opencl_callables():
    """
    Returns an instance of :class:`InKernelCallable` if the function defined by
    *identifier* is known in OpenCL.
    """
    opencl_function_ids = (
            {"dot", "abs",
             "cos", "cosh", "sin", "sinh", "tan", "tanh",
             "acos", "acosh", "asin", "asinh", "atan", "atanh", "atan2",
             "pow", "exp", "log", "log10", "sqrt", "ceil", "floor",
             "max", "min", "fmax", "fmin",
             "fabs",  "erf", "erfc"}
            | set(_CL_SIMPLE_MULTI_ARG_FUNCTIONS)
            | set(VECTOR_LITERAL_FUNCS))

    return {id_: OpenCLCallable(name=id_) for id_ in
        opencl_function_ids}

# }}}


# {{{ symbol mangler

def opencl_symbol_mangler(kernel, name):
    # FIXME: should be more picky about exact names
    if name.startswith("FLT_"):
        return NumpyType(np.dtype(np.float32)), name
    elif name.startswith("DBL_"):
        return NumpyType(np.dtype(np.float64)), name
    elif name.startswith("M_"):
        if name.endswith("_F"):
            return NumpyType(np.dtype(np.float32)), name
        else:
            return NumpyType(np.dtype(np.float64)), name
    elif name == "INFINITY":
        return NumpyType(np.dtype(np.float32)), name
    elif name.startswith("INT_"):
        return NumpyType(np.dtype(np.int32)), name
    elif name.startswith("LONG_"):
        return NumpyType(np.dtype(np.int64)), name
    elif name == "HUGE_VAL":
        return NumpyType(np.dtype(np.float64)), name
    else:
        return None

# }}}


# {{{ preamble generator

def opencl_preamble_generator(preamble_info):
    has_double = False

    for dtype in preamble_info.seen_dtypes:
        if (isinstance(dtype, NumpyType)
                and dtype.numpy_dtype in [np.float64, np.complex128]):
            has_double = True

    if has_double:
        yield ("00_enable_double", """
            #if __OPENCL_C_VERSION__ < 120
            #pragma OPENCL EXTENSION cl_khr_fp64: enable
            #endif
            """)

    from loopy.types import AtomicNumpyType
    seen_64_bit_atomics = any(
            isinstance(dtype, AtomicNumpyType) and dtype.numpy_dtype.itemsize == 8
            for dtype in preamble_info.seen_atomic_dtypes)

    if seen_64_bit_atomics:
        # FIXME: Should gate on "CL1" atomics style
        yield ("00_enable_64bit_atomics", """
            #pragma OPENCL EXTENSION cl_khr_int64_base_atomics : enable
            """)

    from loopy.tools import remove_common_indentation
    kernel = preamble_info.kernel

    idx_ctype = kernel.target.dtype_to_typename(kernel.index_dtype)
    yield ("00_declare_gid_lid",
            remove_common_indentation(f"""
                #define lid(N) (({idx_ctype}) get_local_id(N))
                #define gid(N) (({idx_ctype}) get_group_id(N))
                """))

    for func in preamble_info.seen_functions:
        if func.name == "pow" and func.c_name == "powf32":
            yield ("08_clpowf32", """
                inline float powf32(float x, float y) {
                return pow(x, y);
                }""")

        if func.name == "pow" and func.c_name == "powf64":
            yield ("08_clpowf64", """
                inline double powf64(double x, double y) {
                return pow(x, y);
                }""")

# }}}


# {{{ expression mapper

class ExpressionToOpenCLCExpressionMapper(ExpressionToCExpressionMapper):

    def wrap_in_typecast(self, actual_type, needed_dtype, s):
        if needed_dtype.dtype.kind == "b" and actual_type.dtype.kind == "f":
            # CL does not perform implicit conversion from float-type to a bool.
            from pymbolic.primitives import Comparison
            return Comparison(s, "!=", 0)

        return super().wrap_in_typecast(actual_type, needed_dtype, s)

    def map_group_hw_index(self, expr, type_context):
        return var("gid")(expr.axis)

    def map_local_hw_index(self, expr, type_context):
        return var("lid")(expr.axis)

# }}}


# {{{ target

class OpenCLTarget(CFamilyTarget):
    """A target for the OpenCL C heterogeneous compute programming language.
    """

    def __init__(self, atomics_flavor=None, use_int8_for_bool=True):
        """
        :arg atomics_flavor: one of ``"cl1"`` (C11-style atomics from OpenCL 2.0),
            ``"cl1"`` (OpenCL 1.1 atomics, using bit-for-bit compare-and-swap
            for floating point), ``"cl1-exch"`` (OpenCL 1.1 atomics, using
            double-exchange for floating point--not yet supported).
        :arg use_int8_for_bool: Size of *bool* is undefined as per
            OpenCL spec, if *True* all bool_ variables would be treated
            as int8's.
        """
        super().__init__()

        if atomics_flavor is None:
            atomics_flavor = "cl1"

        if atomics_flavor not in ["cl1", "cl2"]:
            raise ValueError("unsupported atomics flavor: %s" % atomics_flavor)

        self.atomics_flavor = atomics_flavor
        self.use_int8_for_bool = use_int8_for_bool

    def split_kernel_at_global_barriers(self):
        return True

    def get_device_ast_builder(self):
        return OpenCLCASTBuilder(self)

    @memoize_method
    def get_dtype_registry(self):
        from loopy.target.c.compyte.dtypes import (
            DTypeRegistry,
            fill_registry_with_opencl_c_types,
        )

        result = DTypeRegistry()
        fill_registry_with_opencl_c_types(result)

        # no complex number support--needs PyOpenCLTarget

        _register_vector_types(result)

        if self.atomics_flavor == "cl1":
            result = DTypeRegistryWrapperWithCL1Atomics(result)
        else:
            raise NotImplementedError("atomics flavor: %s" % self.atomics_flavor)

        if self.use_int8_for_bool:
            result = DTypeRegistryWrapperWithInt8ForBool(result)

        return result

    def is_vector_dtype(self, dtype):
        return (isinstance(dtype, NumpyType)
                and dtype.numpy_dtype in list(vec.types.values()))

    def vector_dtype(self, base, count):
        return NumpyType(vec.types[base.numpy_dtype, count])

# }}}


# {{{ ast builder

class OpenCLCASTBuilder(CFamilyASTBuilder):
    # {{{ library

    @property
    def known_callables(self):
        callables = super().known_callables
        callables.update(get_opencl_callables())
        return callables

    def symbol_manglers(self):
        return (
                [*super().symbol_manglers(), opencl_symbol_mangler])

    def preamble_generators(self):

        return (
                [*super().preamble_generators(), opencl_preamble_generator])

    # }}}

    # {{{ top-level codegen

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], Generable]:
        preambles, fdecl = super().get_function_declaration(
                codegen_state, codegen_result, schedule_index)

        from loopy.target.c import FunctionDeclarationWrapper
        assert isinstance(fdecl, FunctionDeclarationWrapper)
        if not codegen_state.is_entrypoint:
            # auxiliary kernels need not mention opencl specific qualifiers
            # for a functions signature
            return preambles, fdecl

        return preambles, FunctionDeclarationWrapper(
                self._wrap_kernel_decl(codegen_state, schedule_index, fdecl.subdecl))

    def _wrap_kernel_decl(
            self, codegen_state: CodeGenerationState, schedule_index: int,
            fdecl: Declarator) -> Declarator:
        from cgen.opencl import CLKernel, CLRequiredWorkGroupSize
        fdecl = CLKernel(fdecl)

        from loopy.schedule import get_insn_ids_for_block_at
        assert codegen_state.kernel.linearization is not None
        _, local_sizes = codegen_state.kernel.get_grid_sizes_for_insn_ids_as_exprs(
                get_insn_ids_for_block_at(
                    codegen_state.kernel.linearization, schedule_index),
                codegen_state.callables_table)

        from loopy.symbolic import get_dependencies
        if not get_dependencies(local_sizes):
            # sizes can't have parameter dependencies if they are
            # to be used in static WG size.

            fdecl = CLRequiredWorkGroupSize(local_sizes, fdecl)

        return fdecl

    def generate_top_of_body(self, codegen_state):
        from loopy.kernel.data import ImageArg
        if any(isinstance(arg, ImageArg) for arg in codegen_state.kernel.args):
            from cgen import Const, Initializer, Value
            return [
                    Initializer(Const(Value("sampler_t", "loopy_sampler")),
                        "CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP "
                        "| CLK_FILTER_NEAREST")
                    ]

        return []

    # }}}

    def get_expression_to_c_expression_mapper(self, codegen_state):
        return ExpressionToOpenCLCExpressionMapper(codegen_state)

    def add_vector_access(self, access_expr, index):
        # The 'int' avoids an 'L' suffix for long ints.
        return access_expr.attr("s%s" % hex(int(index))[2:])

    def emit_barrier(self, synchronization_kind, mem_kind, comment):
        """
        :arg kind: ``"local"`` or ``"global"``
        :return: a :class:`loopy.codegen.GeneratedInstruction`.
        """
        if synchronization_kind == "local":
            if comment:
                comment = " /* %s */" % comment

            mem_kind = mem_kind.upper()

            from cgen import Statement
            return Statement(f"barrier(CLK_{mem_kind}_MEM_FENCE){comment}")
        elif synchronization_kind == "global":
            raise LoopyError("OpenCL does not have global barriers")
        else:
            raise LoopyError("unknown barrier kind")

    # {{{ declarators

    def wrap_decl_for_address_space(
            self, decl: Declarator, address_space: AddressSpace) -> Declarator:
        from cgen.opencl import CLGlobal, CLLocal
        if address_space == AddressSpace.GLOBAL:
            return CLGlobal(decl)
        elif address_space == AddressSpace.LOCAL:
            return CLLocal(decl)
        elif address_space == AddressSpace.PRIVATE:
            return decl
        else:
            raise ValueError("unexpected temporary variable address space: %s"
                    % address_space)

    def wrap_global_constant(self, decl: Declarator) -> Declarator:
        from cgen.opencl import CLConstant, CLGlobal
        assert isinstance(decl, CLGlobal)
        decl = decl.subdecl

        return CLConstant(decl)

    # duplicated in CUDA, update there if updating here
    def get_array_base_declarator(self, ary: ArrayBase) -> Declarator:
        dtype = ary.dtype

        vec_size = ary.vector_length()
        if vec_size > 1:
            dtype = self.target.vector_dtype(dtype, vec_size)

        if ary.dim_tags:
            for dim_tag in ary.dim_tags:
                if isinstance(dim_tag, (FixedStrideArrayDimTag, VectorArrayDimTag)):
                    # we're OK with those
                    pass

                else:
                    raise NotImplementedError(
                        f"{type(self).__name__} does not understand axis tag "
                        f"'{type(dim_tag)}.")

        from loopy.target.c import POD
        return POD(self, dtype, ary.name)

    def get_constant_arg_declarator(self, arg: ConstantArg) -> Declarator:
        from cgen import RestrictPointer
        from cgen.opencl import CLConstant

        # constant *is* an address space as far as CL is concerned, do not re-wrap
        return CLConstant(RestrictPointer(self.get_array_base_declarator(
                arg)))

    def get_image_arg_declarator(
            self, arg: ImageArg, is_written: bool) -> Declarator:
        from cgen.opencl import CLImage

        mode: Literal["r", "w"] = "w" if is_written else "r"
        return CLImage(arg.num_target_axes(), mode, arg.name)

    # }}}

    # {{{ atomics

    def emit_atomic_init(self, codegen_state, lhs_atomicity, lhs_var,
            lhs_expr, rhs_expr, lhs_dtype, rhs_type_context):
        # for the CL1 flavor, this is as simple as a regular update with whatever
        # the RHS value is...

        return self.emit_atomic_update(codegen_state, lhs_atomicity, lhs_var,
            lhs_expr, rhs_expr, lhs_dtype, rhs_type_context)

    def emit_unroll_hint(self, value):
        # See https://man.opencl.org/attributes-loopUnroll.html
        from cgen import Line
        if value:
            return Line(f"__attribute__((opencl_unroll_hint({value})))")
        else:
            return Line("__attribute__((opencl_unroll_hint))")

    def emit_atomic_update(self, codegen_state, lhs_atomicity, lhs_var,
            lhs_expr, rhs_expr, lhs_dtype, rhs_type_context):
        from pymbolic.mapper.stringifier import PREC_NONE

        # FIXME: Could detect operations, generate atomic_{add,...} when
        # appropriate.

        if isinstance(lhs_dtype, NumpyType) and lhs_dtype.numpy_dtype in [
                np.int32, np.int64, np.float32, np.float64]:
            from cgen import Assign, Block, DoWhile

            from loopy.target.c import POD
            old_val_var = codegen_state.var_name_generator("loopy_old_val")
            new_val_var = codegen_state.var_name_generator("loopy_new_val")

            from loopy.kernel.data import AddressSpace, TemporaryVariable
            ecm = codegen_state.expression_to_code_mapper.with_assignments(
                    {
                        old_val_var: TemporaryVariable(old_val_var, lhs_dtype,
                            shape=()),
                        new_val_var: TemporaryVariable(new_val_var, lhs_dtype,
                            shape=()),
                        })

            lhs_expr_code = ecm(lhs_expr, prec=PREC_NONE, type_context=None)

            from pymbolic import var
            from pymbolic.mapper.substitutor import make_subst_func

            from loopy.symbolic import SubstitutionMapper

            subst = SubstitutionMapper(
                    make_subst_func({lhs_expr: var(old_val_var)}))
            rhs_expr_code = ecm(subst(rhs_expr), prec=PREC_NONE,
                    type_context=rhs_type_context,
                    needed_dtype=lhs_dtype)

            if lhs_dtype.numpy_dtype.itemsize == 4:
                func_name = "atomic_cmpxchg"
            elif lhs_dtype.numpy_dtype.itemsize == 8:
                func_name = "atom_cmpxchg"
            else:
                raise LoopyError("unexpected atomic size")

            cast_str = ""
            old_val = old_val_var
            new_val = new_val_var

            if lhs_dtype.numpy_dtype.kind == "f":
                if lhs_dtype.numpy_dtype == np.float32:
                    ctype = "int"
                elif lhs_dtype.numpy_dtype == np.float64:
                    ctype = "long"
                else:
                    raise AssertionError()

                from loopy.kernel.data import ArrayArg, TemporaryVariable
                if (
                        isinstance(lhs_var, ArrayArg)
                        and
                        lhs_var.address_space == AddressSpace.GLOBAL):
                    var_kind = "__global"
                elif (
                        isinstance(lhs_var, ArrayArg)
                        and
                        lhs_var.address_space == AddressSpace.LOCAL):
                    var_kind = "__local"
                elif (
                        isinstance(lhs_var, TemporaryVariable)
                        and lhs_var.address_space == AddressSpace.LOCAL):
                    var_kind = "__local"
                elif (
                        isinstance(lhs_var, TemporaryVariable)
                        and lhs_var.address_space == AddressSpace.GLOBAL):
                    var_kind = "__global"
                else:
                    raise LoopyError("unexpected kind of variable '%s' in "
                            "atomic operation: '%s'"
                            % (lhs_var.name, type(lhs_var).__name__))

                old_val = "*(%s *) &" % ctype + old_val
                new_val = "*(%s *) &" % ctype + new_val
                cast_str = f"({var_kind} {ctype} *) "

            return Block([
                POD(self, NumpyType(lhs_dtype.dtype),
                    old_val_var),
                POD(self, NumpyType(lhs_dtype.dtype),
                    new_val_var),
                DoWhile(
                    "%(func_name)s("
                    "%(cast_str)s&(%(lhs_expr)s), "
                    "%(old_val)s, "
                    "%(new_val)s"
                    ") != %(old_val)s"
                    % {
                        "func_name": func_name,
                        "cast_str": cast_str,
                        "lhs_expr": lhs_expr_code,
                        "old_val": old_val,
                        "new_val": new_val,
                        },
                    Block([
                        Assign(old_val_var, lhs_expr_code),
                        Assign(new_val_var, rhs_expr_code),
                        ])
                    )
                ])
        else:
            raise NotImplementedError("atomic update for '%s'" % lhs_dtype)

    # }}}

# }}}


# {{{ volatile mem access target

class VolatileMemExpressionToOpenCLCExpressionMapper(
        ExpressionToOpenCLCExpressionMapper):
    def make_subscript(self, array, base_expr, subscript):
        registry = self.codegen_state.ast_builder.target.get_dtype_registry()

        from loopy.kernel.data import AddressSpace
        if array.address_space == AddressSpace.GLOBAL:
            aspace = "__global "
        elif array.address_space == AddressSpace.LOCAL:
            aspace = "__local "
        elif array.address_space == AddressSpace.PRIVATE:
            aspace = ""
        else:
            raise ValueError("unexpected value of address space")

        from pymbolic import var
        return var(
                "(%s volatile %s *) "
                % (
                    registry.dtype_to_ctype(array.dtype),
                    aspace,
                    )
                )(base_expr)[subscript]


class VolatileMemOpenCLCASTBuilder(OpenCLCASTBuilder):
    def get_expression_to_c_expression_mapper(self, codegen_state):
        return VolatileMemExpressionToOpenCLCExpressionMapper(codegen_state)


class VolatileMemOpenCLTarget(OpenCLTarget):
    def get_device_ast_builder(self):
        return VolatileMemOpenCLCASTBuilder(self)

# }}}

# vim: foldmethod=marker



import numpy as np
import numpy.linalg as la

import pyopencl as cl
import pyopencl.array
import pyopencl.clrandom

import loopy as lp


def main():
    import pathlib
    fn = pathlib.Path(__file__).parent / "matmul.floopy"

    with open(fn) as inf:
        source = inf.read()

    dgemm = lp.parse_transformed_fortran(source, filename=fn)

    ctx = cl.create_some_context()
    queue = cl.CommandQueue(ctx)

    n = 2048
    a = cl.array.empty(queue, (n, n), dtype=np.float64, order="F")
    b = cl.array.empty(queue, (n, n), dtype=np.float64, order="F")
    c = cl.array.zeros(queue, (n, n), dtype=np.float64, order="F")
    cl.clrandom.fill_rand(a)
    cl.clrandom.fill_rand(b)

    dgemm = lp.set_options(dgemm, write_code=True)

    dgemm(queue, a=a, b=b, alpha=1, c=c)

    c_ref = (a.get() @ b.get())
    assert la.norm(c_ref - c.get())/la.norm(c_ref) < 1e-10


if __name__ == "__main__":
    main()



import ctypes
import ctypes.util
import os
from tempfile import TemporaryDirectory
from time import time

import numpy as np
import numpy.linalg as la

import loopy as lp
from loopy.tools import (
    address_from_numpy,
    build_ispc_shared_lib,
    cptr_from_numpy,
    empty_aligned,
)
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


def transform(knl, vars, stream_dtype):
    vars = [v.strip() for v in vars.split(",")]
    knl = lp.assume(knl, "n>0")
    knl = lp.split_iname(
        knl, "i", 2**18, outer_tag="g.0", slabs=(0, 1))
    knl = lp.split_iname(knl, "i_inner", 8, inner_tag="l.0")

    knl = lp.add_and_infer_dtypes(knl, dict.fromkeys(vars, stream_dtype))

    knl = lp.set_argument_order(knl, [*vars, "n"])

    return knl


def gen_code(knl):
    codegen_result = lp.generate_code_v2(knl)

    return codegen_result.device_code() + "\n" + codegen_result.host_code()


NRUNS = 10
ALIGN_TO = 4096
ARRAY_SIZE = 2**28

if 0:
    STREAM_DTYPE = np.float64
    STREAM_CTYPE = ctypes.c_double
else:
    STREAM_DTYPE = np.float32
    STREAM_CTYPE = ctypes.c_float

if 1:
    INDEX_DTYPE = np.int32
    INDEX_CTYPE = ctypes.c_int
else:
    INDEX_DTYPE = np.int64
    INDEX_CTYPE = ctypes.c_longlong


def main():
    this_dir = os.path.dirname(__file__)
    with open(os.path.join(this_dir, "tasksys.cpp")) as ts_file:
        tasksys_source = ts_file.read()

    def make_knl(name, insn, vars):
        knl = lp.make_kernel(
                "{[i]: 0<=i<n}",
                insn,
                target=lp.ISPCTarget(), index_dtype=INDEX_DTYPE,
                name="stream_"+name+"_tasks")

        knl = transform(knl, vars, STREAM_DTYPE)
        return knl

    init_knl = make_knl("init", """
                a[i] = 1
                b[i] = 2
                c[i] = 0
                """, "a,b,c")
    triad_knl = make_knl("triad", """
            a[i] = b[i] + scalar * c[i]
            """, "a,b,c,scalar")

    with TemporaryDirectory() as tmpdir:
        ispc_code = gen_code(init_knl) + gen_code(triad_knl)
        print(ispc_code)

        build_ispc_shared_lib(
                tmpdir,
                [("stream.ispc", ispc_code)],
                [("tasksys.cpp", tasksys_source)],
                cxx_options=["-g", "-fopenmp", "-DISPC_USE_OMP"],
                ispc_options=([
                    # "-g", "--no-omit-frame-pointer",
                    "--target=avx2-i32x8",
                    "--opt=force-aligned-memory",
                    "--opt=disable-loop-unroll",
                    # "--opt=fast-math",
                    # "--opt=disable-fma",
                    ]
                    + (["--addressing=64"] if INDEX_DTYPE == np.int64 else [])
                    ),
                # ispc_bin="/home/andreask/pack/ispc-v1.9.0-linux/ispc",
                quiet=False,
                )

        knl_lib = ctypes.cdll.LoadLibrary(os.path.join(tmpdir, "shared.so"))

        scalar = 5

        a = empty_aligned(ARRAY_SIZE, dtype=STREAM_DTYPE, n=ALIGN_TO)
        b = empty_aligned(ARRAY_SIZE, dtype=STREAM_DTYPE, n=ALIGN_TO)
        c = empty_aligned(ARRAY_SIZE, dtype=STREAM_DTYPE, n=ALIGN_TO)

        print(
                hex(address_from_numpy(a)),
                hex(address_from_numpy(b)),
                hex(address_from_numpy(c)))
        assert address_from_numpy(a) % ALIGN_TO == 0
        assert address_from_numpy(b) % ALIGN_TO == 0
        assert address_from_numpy(c) % ALIGN_TO == 0

        knl_lib.stream_init_tasks(
                cptr_from_numpy(a),
                cptr_from_numpy(b),
                cptr_from_numpy(c),
                INDEX_CTYPE(ARRAY_SIZE),
                )

        def call_kernel():
            knl_lib.stream_triad_tasks(
                    cptr_from_numpy(a),
                    cptr_from_numpy(b),
                    cptr_from_numpy(c),
                    STREAM_CTYPE(scalar),
                    INDEX_CTYPE(ARRAY_SIZE),
                    )

        call_kernel()
        call_kernel()

        start_time = time()

        for _irun in range(NRUNS):
            call_kernel()

        elapsed = time() - start_time

        print(elapsed/NRUNS)

        print(1e-9*3*a.nbytes*NRUNS/elapsed, "GB/s")

        assert la.norm(a-b+scalar*c, np.inf) < np.finfo(STREAM_DTYPE).eps * 10


if __name__ == "__main__":
    main()

# vim: foldmethod=marker



from __future__ import annotations


"""OpenCL target integrated with PyOpenCL."""

__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from typing import TYPE_CHECKING, Any, Sequence, cast
from warnings import warn

import numpy as np
from constantdict import constantdict

import pymbolic.primitives as p
from cgen import (
    Block,
    Collection,
    Const,
    Declarator,
    FunctionBody,
    Generable,
    Initializer,
    Line,
    Pointer,
)
from cgen.opencl import CLGlobal

from loopy.diagnostic import LoopyError, LoopyTypeError
from loopy.kernel.data import (
    ArrayArg,
    ConstantArg,
    ImageArg,
    TemporaryVariable,
    ValueArg,
)
from loopy.kernel.function_interface import ScalarCallable
from loopy.schedule import CallKernel
from loopy.target.opencl import (
    ExpressionToOpenCLCExpressionMapper,
    OpenCLCASTBuilder,
    OpenCLTarget,
)
from loopy.target.python import PythonASTBuilderBase
from loopy.types import NumpyType


logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import genpy
    import pyopencl as cl

    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult
    from loopy.kernel import LoopKernel
    from loopy.target.pyopencl_execution import PyOpenCLExecutor
    from loopy.translation_unit import FunctionIdT, TranslationUnit
    from loopy.typing import Expression


# {{{ pyopencl function scopers

class PyOpenCLCallable(ScalarCallable):
    """
    Records information about the callables which are not covered by
    :class:`loopy.target.opencl.OpenCLCallable`
    """
    def with_types(self, arg_id_to_dtype, callables_table):

        name = self.name

        for id in arg_id_to_dtype:
            # since all the below functions are single arg.
            if not -1 <= id <= 0:
                raise LoopyError(f"{name} can only take one argument")

        if 0 not in arg_id_to_dtype or arg_id_to_dtype[0] is None:
            # the types provided aren't mature enough to specialize the
            # callable
            return (
                    self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                    callables_table)

        dtype = arg_id_to_dtype[0]

        if name in ["real", "imag", "abs"]:
            if dtype.is_complex():
                if dtype.numpy_dtype == np.complex64:
                    tpname = "cfloat"
                elif dtype.numpy_dtype == np.complex128:
                    tpname = "cdouble"
                else:
                    raise LoopyTypeError(f"unexpected complex type '{dtype}'")

                return (
                        self.copy(name_in_target=f"{tpname}_{name}",
                            arg_id_to_dtype=constantdict({
                                0: dtype,
                                -1: NumpyType(np.dtype(dtype.numpy_dtype.type(0).real))
                                })),
                        callables_table)

        if name in ["real", "imag", "conj"]:
            if not dtype.is_complex():
                tpname = dtype.numpy_dtype.type.__name__
                return (
                        self.copy(
                            name_in_target=f"_lpy_{name}_{tpname}",
                            arg_id_to_dtype=constantdict({0: dtype, -1: dtype})),
                        callables_table)

        if name in ["sqrt", "exp", "log",
                "sin", "cos", "tan",
                "sinh", "cosh", "tanh",
                "conj"]:
            if dtype.is_complex():
                # function parameters are complex.
                if dtype.numpy_dtype == np.complex64:
                    tpname = "cfloat"
                elif dtype.numpy_dtype == np.complex128:
                    tpname = "cdouble"
                else:
                    raise LoopyTypeError("unexpected complex type '%s'" % dtype)

                return (
                        self.copy(name_in_target=f"{tpname}_{name}",
                            arg_id_to_dtype=constantdict({0: dtype, -1: dtype})),
                        callables_table)

            # fall back to pure OpenCL for real-valued arguments

        from loopy.target.opencl import OpenCLCallable
        return OpenCLCallable(name,
                arg_id_to_dtype=self.arg_id_to_dtype,
                arg_id_to_descr=self.arg_id_to_descr,
                name_in_target=self.name_in_target).with_types(
                        arg_id_to_dtype, callables_table)

    def generate_preambles(self, target):
        name = self.name_in_target
        if (name.startswith("_lpy_real")
                or name.startswith("_lpy_conj")
                or name.startswith("_lpy_imag")):
            if name.startswith("_lpy_real") or name.startswith("_lpy_conj"):
                ret = "x"
            else:
                ret = "0"

            dtype = self.arg_id_to_dtype[-1]
            ctype = target.dtype_to_typename(dtype)

            yield (f"40_{name}", f"""
                static inline {ctype} {name}({ctype} x) {{
                    return {ret};
                }}
                """)


def get_pyopencl_callables():
    pyopencl_ids = ["sqrt", "exp", "log", "sin", "cos", "tan", "sinh", "cosh",
            "tanh", "conj", "real", "imag", "abs"]
    return {id_: PyOpenCLCallable(name=id_) for id_ in pyopencl_ids}

# }}}


# {{{ preamble generator

def pyopencl_preamble_generator(preamble_info):
    has_double = False
    has_complex = False

    from loopy.types import NumpyType
    for dtype in preamble_info.seen_dtypes:
        if (isinstance(dtype, NumpyType)
                and dtype.dtype in [np.float64, np.complex128]):
            has_double = True
        if dtype.involves_complex():
            has_complex = True

    if has_complex:
        if has_double:
            yield ("10_include_complex_header", """
                #define PYOPENCL_DEFINE_CDOUBLE
                #ifndef PYOPENCL_COMPLEX_ENABLE_EXTENDED_ALIGNMENT
                #define PYOPENCL_COMPLEX_ENABLE_EXTENDED_ALIGNMENT 1
                #endif

                #include <pyopencl-complex.h>
                """)
        else:
            yield ("10_include_complex_header", """
                #ifndef PYOPENCL_COMPLEX_ENABLE_EXTENDED_ALIGNMENT
                #define PYOPENCL_COMPLEX_ENABLE_EXTENDED_ALIGNMENT 1
                #endif

                #include <pyopencl-complex.h>
                """)

# }}}


# {{{ expression mapper

class ExpressionToPyOpenCLCExpressionMapper(ExpressionToOpenCLCExpressionMapper):
    def complex_type_name(self, dtype):
        from loopy.types import NumpyType
        if not isinstance(dtype, NumpyType):
            raise LoopyError("'%s' is not a complex type" % dtype)

        if dtype.dtype == np.complex64:
            return "cfloat"
        if dtype.dtype == np.complex128:
            return "cdouble"
        else:
            raise RuntimeError

    def wrap_in_typecast(self, actual_type, needed_type, s):
        if (actual_type.is_complex() and needed_type.is_complex()
                and actual_type != needed_type):
            return p.Variable("%s_cast" % self.complex_type_name(needed_type))(s)
        elif not actual_type.is_complex() and needed_type.is_complex():
            return p.Variable("%s_fromreal" % self.complex_type_name(needed_type))(
                    s)
        else:
            return super().wrap_in_typecast(actual_type, needed_type, s)

    def map_sum(self, expr, type_context):
        # I've added 'type_context == "i"' because of the following
        # idiotic corner case: Code generation for subscripts comes
        # through here, and it may involve variables that we know
        # nothing about (offsets and such). If we fall into the allow_complex
        # branch, we'll try to do type inference on these variables,
        # and stuff breaks. This band-aid works around that. -AK
        if not self.allow_complex or type_context == "i":
            return super().map_sum(expr, type_context)

        tgt_dtype = self.infer_type(expr)
        is_complex = tgt_dtype.is_complex()

        if not is_complex:
            return super().map_sum(expr, type_context)
        elif not self.kernel.options.allow_fp_reordering:
            if len(expr.children) == 0:
                return tgt_dtype(0)

            tgt_name = self.complex_type_name(tgt_dtype)
            result = None
            lhs_is_complex = False

            for child in expr.children:
                rhs_is_complex = self.infer_type(child).is_complex()
                if rhs_is_complex:
                    child_val = self.rec(child, type_context, tgt_dtype)
                else:
                    child_val = self.rec(child, type_context)

                if result is None:
                    result = child_val
                elif lhs_is_complex and rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_add")(result, child_val)
                elif lhs_is_complex and not rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_addr")(result, child_val)
                elif not lhs_is_complex and rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_radd")(result, child_val)
                else:
                    result = p.Sum((result, child_val))
                lhs_is_complex = lhs_is_complex or rhs_is_complex
            return result
        else:
            tgt_name = self.complex_type_name(tgt_dtype)

            reals = []
            complexes = []
            for child in expr.children:
                if self.infer_type(child).is_complex():
                    complexes.append(child)
                else:
                    reals.append(child)

            real_sum = p.flattened_sum([self.rec(r, type_context) for r in reals])

            c_applied = [self.rec(c, type_context, tgt_dtype) for c in complexes]

            mul_name = f"{tgt_name}_mul"

            def binary_tree_add(start, end):
                if start + 1 == end:
                    return c_applied[start]
                mid = (start + end)//2
                lsum = binary_tree_add(start, mid)
                rsum = binary_tree_add(mid, end)

                # FMAs should ideally be recognized by the compiler, but some
                # compilers fail to do so. For eg:
                #
                #    res = complex_add(c, complex_mul(a, b))
                #
                # leads to code that looks like below because of the temporary
                # given by ``complex_mul(a, b)``.
                #
                #    tmp.real = a.real * b.real - a.imag * b.imag
                #    tmp.imag = a.real * b.imag + a.imag * b.real
                #    res.real = c.real + tmp.real
                #    res.imag = c.imag + tmp.imag
                #
                # clang can fuse across multiple statements like this with
                # -ffp-contract=fast which is the default for PTX codegen, but
                # for some unknown reason, clang fails to see the FMAs.
                #
                # We need to do this only for complex as we have temporaries
                # only in complex. For reals, the code generated looks like
                #
                #    res = c + a * b
                #
                # and clang is able to generate an FMA for this code.

                if isinstance(lsum, p.Call) and isinstance(lsum.function,
                        p.Variable) and lsum.function.name == mul_name:
                    return p.Variable(f"{tgt_name}_fma")(*lsum.parameters, rsum)

                elif isinstance(rsum, p.Call) and isinstance(rsum.function,
                        p.Variable) and rsum.function.name == mul_name:
                    return p.Variable(f"{tgt_name}_fma")(*rsum.parameters, lsum)

                else:
                    return p.Variable(f"{tgt_name}_add")(lsum, rsum)

            complex_sum = binary_tree_add(0, len(c_applied))

            if reals:
                return p.Variable(f"{tgt_name}_radd")(real_sum, complex_sum)
            else:
                return complex_sum

    def map_product(self, expr, type_context):
        # I've added 'type_context == "i"' because of the following
        # idiotic corner case: Code generation for subscripts comes
        # through here, and it may involve variables that we know
        # nothing about (offsets and such). If we fall into the allow_complex
        # branch, we'll try to do type inference on these variables,
        # and stuff breaks. This band-aid works around that. -AK
        if not self.allow_complex or type_context == "i":
            return super().map_product(expr, type_context)

        tgt_dtype = self.infer_type(expr)
        is_complex = tgt_dtype.is_complex()

        if not is_complex:
            return super().map_product(expr, type_context)
        elif not self.kernel.options.allow_fp_reordering:
            tgt_name = self.complex_type_name(tgt_dtype)

            result = None
            lhs_is_complex = False
            for child in expr.children:
                rhs_is_complex = self.infer_type(child).is_complex()
                if rhs_is_complex:
                    child_val = self.rec(child, type_context, tgt_dtype)
                else:
                    child_val = self.rec(child, type_context)

                if result is None:
                    result = child_val
                elif lhs_is_complex and rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_mul")(result, child_val)
                elif lhs_is_complex and not rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_mulr")(result, child_val)
                elif not lhs_is_complex and rhs_is_complex:
                    result = p.Variable(f"{tgt_name}_rmul")(result, child_val)
                else:
                    result = p.Product((result, child_val))
                lhs_is_complex = lhs_is_complex or rhs_is_complex
            return result
        else:
            tgt_name = self.complex_type_name(tgt_dtype)

            reals = []
            complexes = []
            for child in expr.children:
                if self.infer_type(child).is_complex():
                    complexes.append(child)
                else:
                    reals.append(child)

            real_prd = p.flattened_product(
                    [self.rec(r, type_context) for r in reals])

            c_applied = [self.rec(c, type_context, tgt_dtype) for c in complexes]

            def binary_tree_mul(start, end):
                if start + 1 == end:
                    return c_applied[start]
                mid = (start + end)//2
                lsum = binary_tree_mul(start, mid)
                rsum = binary_tree_mul(mid, end)
                return p.Variable("%s_mul" % tgt_name)(lsum, rsum)

            complex_prd = binary_tree_mul(0, len(complexes))

            if reals:
                return p.Variable("%s_rmul" % tgt_name)(real_prd, complex_prd)
            else:
                return complex_prd

    def map_quotient(self, expr, type_context):
        n_dtype = self.infer_type(expr.numerator).numpy_dtype
        d_dtype = self.infer_type(expr.denominator).numpy_dtype
        tgt_dtype = self.infer_type(expr)
        n_complex = "c" == n_dtype.kind
        d_complex = "c" == d_dtype.kind

        if not self.allow_complex or (not (n_complex or d_complex)):
            return super().map_quotient(expr, type_context)

        if n_complex and not d_complex:
            return p.Variable("%s_divider" % self.complex_type_name(tgt_dtype))(
                    self.rec(expr.numerator, type_context, tgt_dtype),
                    self.rec(expr.denominator, type_context))
        elif not n_complex and d_complex:
            return p.Variable("%s_rdivide" % self.complex_type_name(tgt_dtype))(
                    self.rec(expr.numerator, type_context),
                    self.rec(expr.denominator, type_context, tgt_dtype))
        else:
            return p.Variable("%s_divide" % self.complex_type_name(tgt_dtype))(
                    self.rec(expr.numerator, type_context, tgt_dtype),
                    self.rec(expr.denominator, type_context, tgt_dtype))

    def map_constant(self, expr, type_context):
        if isinstance(expr, (complex, np.complexfloating)):
            try:
                dtype = expr.dtype
            except AttributeError:
                # (COMPLEX_GUESS_LOGIC) This made it through type 'guessing' in
                # type inference, and it was concluded there (search for
                # COMPLEX_GUESS_LOGIC in loopy.type_inference), that no
                # accuracy was lost by using single precision.
                cast_type = "cfloat"
            else:
                if dtype == np.complex128:
                    cast_type = "cdouble"
                elif dtype == np.complex64:
                    cast_type = "cfloat"
                else:
                    raise RuntimeError("unsupported complex type in expression "
                            "generation: %s" % type(expr))

            return p.Variable("%s_new" % cast_type)(self.rec(expr.real,
                                                             type_context),
                                                    self.rec(expr.imag,
                                                             type_context))

        return super().map_constant(expr, type_context)

    def map_power(self, expr, type_context):
        tgt_dtype = self.infer_type(expr)
        base_dtype = self.infer_type(expr.base)
        exponent_dtype = self.infer_type(expr.exponent)

        if not self.allow_complex or (not tgt_dtype.is_complex()):
            return super().map_power(expr, type_context)

        if expr.exponent in [2, 3, 4]:
            value = expr.base
            for _i in range(expr.exponent-1):
                value = value * expr.base
            return self.rec(value, type_context)
        else:
            b_complex = base_dtype.is_complex()
            e_complex = exponent_dtype.is_complex()

            if b_complex and not e_complex:
                return p.Variable("%s_powr" % self.complex_type_name(tgt_dtype))(
                        self.rec(expr.base, type_context, tgt_dtype),
                        self.rec(expr.exponent, type_context))
            else:
                return p.Variable("%s_pow" % self.complex_type_name(tgt_dtype))(
                        self.rec(expr.base, type_context, tgt_dtype),
                        self.rec(expr.exponent, type_context, tgt_dtype))

# }}}


# {{{ target

class PyOpenCLTarget(OpenCLTarget):
    """A code generation target that takes special advantage of :mod:`pyopencl`
    features such as run-time knowledge of the target device (to generate
    warnings) and support for complex numbers.
    """

    # FIXME make prefixes conform to naming rules
    # (see Reference: Loopy's Model of a Kernel)

    host_program_name_prefix = "_lpy_host_"
    host_program_name_suffix = ""

    # FIXME Not yet complete
    limit_arg_size_nbytes: int | None
    pointer_size_nbytes: int

    def __init__(
            self, device=None, *, pyopencl_module_name: str = "_lpy_cl",
            atomics_flavor=None, use_int8_for_bool: bool = True,
            limit_arg_size_nbytes: int | None = None,
            pointer_size_nbytes: int | None = None
            ) -> None:
        # This ensures the dtype registry is populated.
        import pyopencl.tools

        super().__init__(
            atomics_flavor=atomics_flavor,
            use_int8_for_bool=use_int8_for_bool)

        import pyopencl.version
        if pyopencl.version.VERSION < (2021, 2):
            raise RuntimeError("The version of loopy you have installed "
                    "generates invoker code that requires PyOpenCL 2021.2 "
                    "or newer.")

        if device is not None:
            warn("Passing device is deprecated, it will stop working in 2022.",
                    DeprecationWarning, stacklevel=2)

        self.pyopencl_module_name = pyopencl_module_name

        if pointer_size_nbytes is None:
            pointer_size_nbytes = tuple.__itemsize__

        self.limit_arg_size_nbytes = limit_arg_size_nbytes
        self.pointer_size_nbytes = pointer_size_nbytes

    @property
    def device(self):
        warn("PyOpenCLTarget.device is deprecated, it will stop working in 2022.",
                DeprecationWarning, stacklevel=2)
        return None

    # NB: Not including 'device', as that is handled specially here.
    hash_fields = (*OpenCLTarget.hash_fields, "pyopencl_module_name")
    comparison_fields = (*OpenCLTarget.comparison_fields, "pyopencl_module_name")

    def get_host_ast_builder(self):
        return PyOpenCLPythonASTBuilder(self)

    def get_device_ast_builder(self):
        return PyOpenCLCASTBuilder(self)

    # {{{ types

    def get_dtype_registry(self):
        from pyopencl.compyte.dtypes import TYPE_REGISTRY
        result = TYPE_REGISTRY

        from loopy.target.opencl import (
            DTypeRegistryWrapperWithCL1Atomics,
            DTypeRegistryWrapperWithInt8ForBool,
        )

        result = DTypeRegistryWrapperWithInt8ForBool(result)
        if self.atomics_flavor == "cl1":
            result = DTypeRegistryWrapperWithCL1Atomics(result)
        else:
            raise NotImplementedError("atomics flavor: %s" % self.atomics_flavor)

        return result

    def is_vector_dtype(self, dtype):
        try:
            import pyopencl.cltypes as cltypes
            vec_types = cltypes.vec_types
        except ImportError:
            from pyopencl.array import vec
            vec_types = vec.types

        return (isinstance(dtype, NumpyType)
                and dtype.numpy_dtype in list(vec_types.values()))

    def vector_dtype(self, base, count):
        try:
            import pyopencl.cltypes as cltypes
            vec_types = cltypes.vec_types
        except ImportError:
            from pyopencl.array import vec
            vec_types = vec.types

        return NumpyType(vec_types[base.numpy_dtype, count])

    def alignment_requirement(self, type_decl):
        import struct

        fmt = (type_decl.struct_format()
                .replace("F", "ff")
                .replace("D", "dd"))

        return struct.calcsize(fmt)

    # }}}

    def get_kernel_executor_cache_key(self, queue, **kwargs):
        return (queue.context, kwargs["entrypoint"])

    # type-ignore because we're making things from *args: Any more concrete,
    # and mypy doesn't like it.
    def get_kernel_executor(self, t_unit: TranslationUnit,  # type: ignore[override]
                            queue_or_context: cl.CommandQueue | cl.Context,
                            *args: Any, entrypoint: FunctionIdT, **kwargs: Any
                            ) -> PyOpenCLExecutor:
        from pyopencl import CommandQueue
        if isinstance(queue_or_context, CommandQueue):
            context = queue_or_context.context
        else:
            context = queue_or_context

        from loopy.target.pyopencl_execution import PyOpenCLExecutor
        return PyOpenCLExecutor(context, t_unit, entrypoint=entrypoint)

# }}}


# {{{ host code: value arg setup

def generate_value_arg_setup(
        kernel: LoopKernel, passed_names: Sequence[str]
        ) -> genpy.Suite:
    options = kernel.options

    from genpy import If, Raise, Statement as S, Suite

    import loopy as lp
    from loopy.kernel.array import ArrayBase

    result: list[genpy.Generable] = []
    gen = result.append

    buf_indices_and_args = []
    buf_pack_indices_and_args = []

    from pyopencl.invoker import BUF_PACK_TYPECHARS

    def add_buf_arg(arg_idx, typechar, expr_str):
        if typechar in BUF_PACK_TYPECHARS:
            buf_pack_indices_and_args.append(arg_idx)
            buf_pack_indices_and_args.append(repr(typechar.encode()))
            buf_pack_indices_and_args.append(expr_str)
        else:
            buf_indices_and_args.append(arg_idx)
            buf_indices_and_args.append(f"pack('{typechar}', {expr_str})")

    for arg_idx, passed_name in enumerate(passed_names):
        if passed_name in kernel.all_inames():
            add_buf_arg(arg_idx, kernel.index_dtype.numpy_dtype.char, passed_name)
            continue

        var_descr = kernel.get_var_descriptor(passed_name)
        assert var_descr.dtype is not None

        if not isinstance(var_descr, lp.ValueArg):
            assert isinstance(var_descr, ArrayBase)

            continue

        if not options.skip_arg_checks:
            gen(If(f"{passed_name} is None",
                Raise('RuntimeError("input argument \'{var_descr.name}\' '
                        'must be supplied")')))

        if var_descr.dtype.is_composite():
            buf_indices_and_args.append(arg_idx)
            buf_indices_and_args.append(f"{passed_name}")

        elif var_descr.dtype.is_complex():
            assert isinstance(var_descr.dtype, NumpyType)

            dtype = var_descr.dtype

            if dtype.numpy_dtype == np.complex64:
                arg_char = "f"
            elif dtype.numpy_dtype == np.complex128:
                arg_char = "d"
            else:
                raise TypeError("unexpected complex type: %s" % dtype)

            buf_indices_and_args.append(arg_idx)
            buf_indices_and_args.append(
                f"_lpy_pack('{arg_char}{arg_char}', "
                f"{passed_name}.real, {passed_name}.imag)")

        elif isinstance(var_descr.dtype, NumpyType):
            add_buf_arg(arg_idx, var_descr.dtype.dtype.char, passed_name)

        else:
            raise LoopyError("do not know how to pass argument of type '%s'"
                    % var_descr.dtype)

    for arg_kind, args_and_indices, entry_length in [
            ("_buf", buf_indices_and_args, 2),
            ("_buf_pack", buf_pack_indices_and_args, 3),
            ]:
        assert len(args_and_indices) % entry_length == 0
        if args_and_indices:
            gen(S(f"_lpy_knl._set_arg{arg_kind}_multi("
                    f"({', '.join(str(i) for i in args_and_indices)},), "
                    ")"))

    return Suite(result)

# }}}


def generate_array_arg_setup(
        kernel: LoopKernel, passed_names: Sequence[str],
        ) -> genpy.Generable:
    from genpy import Statement as S, Suite

    from loopy.kernel.array import ArrayBase

    result: list[genpy.Generable] = []
    gen = result.append

    cl_indices_and_args: list[int | str] = []
    for arg_idx, passed_name in enumerate(passed_names):
        if passed_name in kernel.all_inames():
            continue

        var_descr = kernel.get_var_descriptor(passed_name)
        if isinstance(var_descr, ArrayBase):
            cl_indices_and_args.append(arg_idx)
            cl_indices_and_args.append(passed_name)

    if cl_indices_and_args:
        assert len(cl_indices_and_args) % 2 == 0

        gen(S(f"_lpy_knl._set_arg_multi("
            f"({', '.join(str(i) for i in cl_indices_and_args)},)"
            ")"))

    return Suite(result)


# {{{ host ast builder

class PyOpenCLPythonASTBuilder(PythonASTBuilderBase):
    """A Python host AST builder for integration with PyOpenCL.
    """

    # {{{ code generation guts

    def get_function_definition(
            self, codegen_state, codegen_result,
            schedule_index: int, function_decl, function_body: genpy.Generable
            ) -> genpy.Function:
        assert schedule_index == 0

        from loopy.schedule.tools import get_kernel_arg_info
        kai = get_kernel_arg_info(codegen_state.kernel)

        args = (
                ["_lpy_cl_kernels", "queue", *kai.passed_arg_names,
                    "wait_for=None", "allocator=None"])

        from genpy import For, Function, Line, Return, Statement as S, Suite
        return Function(
                codegen_result.current_program(codegen_state).name,
                args,
                Suite([
                    Line(),
                    ] + [
                    Line(),
                    function_body,
                    Line(),
                    ] + ([
                        For("_tv", "_global_temporaries",
                            # Free global temporaries.
                            # Zero-size temporaries allocate as None, tolerate that.
                            # https://documen.tician.de/pyopencl/tools.html#pyopencl.tools.ImmediateAllocator
                            S("if _tv is not None: _tv.release()"))
                        ] if self._get_global_temporaries(codegen_state) else []
                    ) + [
                    Line(),
                    Return("_lpy_evt"),
                    ]))

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], genpy.Generable | None]:
        # no such thing in Python
        return [], None

    def _get_global_temporaries(self, codegen_state):
        from loopy.kernel.data import AddressSpace

        return sorted(
            (tv for tv in codegen_state.kernel.temporary_variables.values()
            if tv.address_space == AddressSpace.GLOBAL),
            key=lambda tv: tv.name)

    def get_temporary_decls(self, codegen_state, schedule_index):
        from genpy import Assign, Comment, Line
        from pymbolic.mapper.stringifier import PREC_NONE
        ecm = self.get_expression_to_code_mapper(codegen_state)

        global_temporaries = self._get_global_temporaries(codegen_state)
        if not global_temporaries:
            return []

        allocated_var_names = []
        code_lines = []
        code_lines.append(Line())
        code_lines.append(Comment("{{{ allocate global temporaries"))
        code_lines.append(Line())

        for tv in global_temporaries:
            if not tv.base_storage:
                if tv.nbytes:
                    # NB: This does not prevent all zero-size allocations,
                    # as sizes are parametric, and allocation size
                    # could turn out to be zero at runtime.
                    nbytes_str = ecm(tv.nbytes, PREC_NONE, "i")
                    allocated_var_names.append(tv.name)
                    code_lines.append(Assign(tv.name,
                                             f"allocator({nbytes_str})"))
                else:
                    code_lines.append(Assign(tv.name, "None"))

        code_lines.append(Assign("_global_temporaries", "[{tvs}]".format(
            tvs=", ".join(tv for tv in allocated_var_names))))

        code_lines.append(Line())
        code_lines.append(Comment("}}}"))
        code_lines.append(Line())

        return code_lines

    def get_kernel_call(
            self, codegen_state: CodeGenerationState,
            subkernel_name: str,
            gsize: tuple[Expression, ...], lsize: tuple[Expression, ...]
            ) -> genpy.Suite:
        from genpy import Assert, Assign, Comment, Line, Suite

        kernel = codegen_state.kernel

        from loopy.schedule.tools import get_subkernel_arg_info
        skai = get_subkernel_arg_info(kernel, subkernel_name)

        ecm = self.get_expression_to_code_mapper(codegen_state)

        if not gsize:
            gsize = (1,)
        if not lsize:
            lsize = (1,)

        assert isinstance(kernel.target, PyOpenCLTarget)
        regular_arg_names, struct_overflow_arg_names = split_args_for_overflow(
                kernel, skai.passed_names,
                limit_arg_size_nbytes=kernel.target.limit_arg_size_nbytes,
                pointer_size_nbytes=kernel.target.pointer_size_nbytes)

        value_arg_code = generate_value_arg_setup(
                codegen_state.kernel, regular_arg_names)
        array_arg_code = generate_array_arg_setup(
                codegen_state.kernel, regular_arg_names)

        if struct_overflow_arg_names:
            regular_arg_names_set = frozenset(regular_arg_names)
            struct_overflow_arg_names_set = frozenset(
                    struct_overflow_arg_names)

            py_passed_args = []
            struct_pack_types: list[str] = []
            struct_pack_args = []

            for arg_name in skai.passed_names:
                if arg_name in regular_arg_names_set:
                    py_passed_args.append(arg_name)
                else:
                    assert arg_name in struct_overflow_arg_names_set

                    arg = kernel.get_var_descriptor(arg_name)
                    assert arg.dtype is not None
                    if isinstance(arg, ValueArg):
                        struct_pack_types.append(arg.dtype.numpy_dtype.char)
                        struct_pack_args.append(arg_name)
                    elif isinstance(arg, (ArrayArg, ConstantArg, TemporaryVariable)):
                        struct_pack_types.append("P")
                        struct_pack_args.append(f"{arg_name}.svm_ptr")
                    elif isinstance(arg, ImageArg):
                        raise AssertionError()
                    else:
                        raise ValueError(f"unrecognized arg type: '{type(arg)}'")

            cl_arg_count = len(regular_arg_names)
            overflow_args_code = Suite([
                # It's important for _lpy_overflow_args_buf to be in a variable.
                # Otherwise, no reference to it will survive until the kernel
                # launch and the buffer may be released.
                Assign("_lpy_overflow_args_buf",
                    "_lpy_cl.Buffer(queue.context, "
                    "_lpy_cl.mem_flags.READ_ONLY "
                    "| _lpy_cl.mem_flags.COPY_HOST_PTR, "
                    "hostbuf="
                    f"_lpy_pack({''.join(struct_pack_types)!r}, "
                    f"{', '.join(struct_pack_args)}))"),
                Line(f"_lpy_knl.set_arg({cl_arg_count}, _lpy_overflow_args_buf)")
                ])

            cl_arg_count += 1

        else:
            cl_arg_count = len(skai.passed_names)
            overflow_args_code = Suite([])

        import pyopencl.version as cl_ver
        from pymbolic.mapper.stringifier import PREC_NONE
        if cl_ver.VERSION < (2020, 2):
            from warnings import warn
            warn("Your kernel invocation will likely fail because your "
                    "version of PyOpenCL does not support allow_empty_ndrange. "
                    "Please upgrade to version 2020.2 or newer.", stacklevel=2)

        # TODO: Generate finer-grained dependency structure
        return Suite([
            Comment("{{{ enqueue %s" % subkernel_name),
            Line(),
            Assign("_lpy_knl", "_lpy_cl_kernels."+subkernel_name),
            Assert(f"_lpy_knl.num_args == {cl_arg_count}, "
                   f"f'Kernel \"{subkernel_name}\" "
                   f"invoker argument count ({cl_arg_count}) does not match the "
                   # No f"" here since {_lpy_knl.num_args} needs to be evaluated
                   # at runtime, not here.
                   "argument count of the kernel ({_lpy_knl.num_args}).'"),
            Line(),
            value_arg_code,
            array_arg_code,
            overflow_args_code,
            Assign("_lpy_evt",
                   f"{self.target.pyopencl_module_name}.enqueue_nd_range_kernel("
                   "queue, _lpy_knl, "
                   f"{ecm(gsize, prec=PREC_NONE, type_context='i')}, "
                   f"{ecm(lsize, prec=PREC_NONE, type_context='i')}, "
                   # using positional args because pybind is slow with kwargs
                   "None, "  # offset
                   "wait_for, "
                   "True, "  # g_times_l
                   "True, "  # allow_empty_ndrange
                   ")"),
            Assign("wait_for", "[_lpy_evt]"),
            Line(),
            Comment("}}}"),
            Line(),
            ])

    # }}}

# }}}


# {{{ split_args_for_overflow

def split_args_for_overflow(
        kernel: LoopKernel, passed_names: Sequence[str],
        *, limit_arg_size_nbytes: int | None, pointer_size_nbytes: int
        ) -> tuple[Sequence[str], Sequence[str]]:
    if limit_arg_size_nbytes is None:
        return passed_names, []

    regular_arg_names = []
    overflow_arg_names = []

    # Consider that the pointer to the arg overflow struct also occupies
    # argument space.
    running_arg_size = pointer_size_nbytes

    for arg_name in passed_names:
        arg = kernel.get_var_descriptor(arg_name)
        if isinstance(arg, (ValueArg, ArrayArg, ConstantArg, TemporaryVariable)):
            if isinstance(arg, ValueArg):
                assert arg.dtype is not None
                arg_size = arg.dtype.numpy_dtype.itemsize
            else:
                arg_size = pointer_size_nbytes

            if running_arg_size + arg_size > limit_arg_size_nbytes:
                overflow_arg_names.append(arg_name)
            else:
                regular_arg_names.append(arg_name)

            running_arg_size += arg_size

        elif isinstance(arg, ImageArg):
            regular_arg_names.append(arg_name)
        else:
            raise ValueError(f"unrecognized arg type: '{type(arg)}'")

    return regular_arg_names, overflow_arg_names

# }}}


# {{{ device ast builder

class PyOpenCLCASTBuilder(OpenCLCASTBuilder):
    """A C device AST builder for integration with PyOpenCL.
    """

    # {{{ function decl/def, with arg overflow handling

    def get_function_definition(
            self,
            codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult,
            schedule_index: int,
            function_decl: Generable,
            function_body: Generable,
            ) -> Generable:
        assert isinstance(function_body, Block)
        kernel = codegen_state.kernel
        assert kernel.linearization is not None

        subkernel_name = cast("CallKernel",
                kernel.linearization[schedule_index]).kernel_name

        result = []

        from loopy.kernel.data import AddressSpace
        # We only need to write declarations for global variables with
        # the first device program. `is_first_dev_prog` determines
        # whether this is the first device program in the schedule.
        is_first_dev_prog = codegen_state.is_generating_device_code
        for i in range(schedule_index):
            if isinstance(kernel.linearization[i], CallKernel):
                is_first_dev_prog = False
                break
        if is_first_dev_prog:
            for tv in sorted(
                    kernel.temporary_variables.values(),
                    key=lambda key_tv: key_tv.name):

                if tv.address_space == AddressSpace.GLOBAL and (
                        tv.initializer is not None):
                    assert tv.read_only

                    decl = self.wrap_global_constant(
                            self.get_temporary_var_declarator(codegen_state, tv))

                    if tv.initializer is not None:
                        from loopy.target.c import generate_array_literal
                        init_decl = Initializer(decl, generate_array_literal(
                            codegen_state, tv, tv.initializer))
                    else:
                        init_decl = decl

                    result.append(init_decl)

        # {{{ unpack overflow args

        if codegen_state.is_entrypoint:
            from loopy.schedule.tools import get_subkernel_arg_info
            skai = get_subkernel_arg_info(kernel, subkernel_name)

            _, struct_overflow_arg_names = split_args_for_overflow(
                    kernel, skai.passed_names,
                    limit_arg_size_nbytes=self.target.limit_arg_size_nbytes,
                    pointer_size_nbytes=self.target.pointer_size_nbytes)

            arg_unpack_code = [
                    Initializer(
                        self.arg_to_cgen_declarator(
                            kernel, arg_name,
                            is_written=arg_name in skai.written_names),
                        f"_lpy_overflow_args->{arg_name}")
                    for arg_name in struct_overflow_arg_names
                    ] + ([Line()] if struct_overflow_arg_names else [])

            function_body = Block(arg_unpack_code + function_body.contents)

        # }}}

        from loopy.target.c import FunctionDeclarationWrapper

        assert isinstance(function_decl, FunctionDeclarationWrapper)
        if not isinstance(function_body, Block):
            function_body = Block([function_body])

        fbody = FunctionBody(function_decl, function_body)
        if not result:
            return fbody
        else:
            return Collection([*result, Line(), fbody])

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], Generable]:
        kernel = codegen_state.kernel

        assert codegen_state.kernel.linearization is not None
        subkernel_name = cast(
                        "CallKernel",
                        codegen_state.kernel.linearization[schedule_index]
                        ).kernel_name

        from cgen import FunctionDeclaration, Struct, Value

        name_str = codegen_result.current_program(codegen_state).name
        if self.target.fortran_abi:
            name_str += "_"

        from loopy.target.c import FunctionDeclarationWrapper

        if codegen_state.is_entrypoint:
            name = Value("void", name_str)

            # subkernel launches occur only as part of entrypoint kernels for now
            from loopy.schedule.tools import get_subkernel_arg_info
            skai = get_subkernel_arg_info(kernel, subkernel_name)
            passed_names = skai.passed_names
            written_names = skai.written_names

            regular_arg_names, struct_overflow_arg_names = split_args_for_overflow(
                    kernel, passed_names,
                    limit_arg_size_nbytes=self.target.limit_arg_size_nbytes,
                    pointer_size_nbytes=self.target.pointer_size_nbytes)

            arg_overflow_struct_name = f"_lpy_arg_struct_{subkernel_name}"
            arg_overflow_struct = Struct(
                    arg_overflow_struct_name, [
                        self.arg_to_cgen_declarator(
                            kernel, arg_name,
                            is_written=arg_name in written_names)
                        for arg_name in struct_overflow_arg_names])

            if struct_overflow_arg_names:
                logger.info(f"overflowing arguments into SVM buffer: "
                        f"{len(regular_arg_names)} regular/"
                        f"{len(struct_overflow_arg_names)} in buffer "
                        f"for '{subkernel_name}'")
                arg_struct_preambles = [
                        (f"declare-{arg_overflow_struct_name}",
                            str(arg_overflow_struct))
                        ] if struct_overflow_arg_names else []
                arg_struct_args: list[Declarator] = [CLGlobal(Const(Pointer(Value(
                                f"struct {arg_overflow_struct_name}",
                                "_lpy_overflow_args"))))]
            else:
                arg_struct_preambles = []
                arg_struct_args = []

            return arg_struct_preambles, FunctionDeclarationWrapper(
                    self._wrap_kernel_decl(
                        codegen_state, schedule_index,
                        FunctionDeclaration(
                            name,
                            [self.arg_to_cgen_declarator(
                                kernel, arg_name,
                                is_written=arg_name in written_names)
                                for arg_name in regular_arg_names]
                            + arg_struct_args
                            )))
        else:
            name = Value("static void", name_str)
            passed_names = [arg.name for arg in kernel.args]
            written_names = kernel.get_written_variables()

            return [], FunctionDeclarationWrapper(
                    FunctionDeclaration(
                        name,
                        [self.arg_to_cgen_declarator(
                                kernel, arg_name,
                                is_written=arg_name in written_names)
                            for arg_name in passed_names]))
    # }}}

    # {{{ library

    @property
    def known_callables(self):
        from loopy.library.random123 import get_random123_callables

        # order matters: e.g. prefer our abs() over that of the
        # superclass
        callables = super().known_callables
        callables.update(get_pyopencl_callables())
        callables.update(get_random123_callables(self.target))
        return callables

    def preamble_generators(self):
        return ([pyopencl_preamble_generator, *super().preamble_generators()])

    # }}}

    def get_expression_to_c_expression_mapper(self, codegen_state):
        return ExpressionToPyOpenCLCExpressionMapper(codegen_state)

# }}}


# {{{ volatile mem access target

class VolatileMemPyOpenCLCASTBuilder(PyOpenCLCASTBuilder):
    def get_expression_to_c_expression_mapper(self, codegen_state):
        from loopy.target.opencl import VolatileMemExpressionToOpenCLCExpressionMapper
        return VolatileMemExpressionToOpenCLCExpressionMapper(codegen_state)


class VolatileMemPyOpenCLTarget(PyOpenCLTarget):
    def get_device_ast_builder(self):
        return VolatileMemPyOpenCLCASTBuilder(self)

# }}}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2009 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import re

from loopy.diagnostic import LoopyError
from loopy.symbolic import FortranDivision, IdentityMapper


class DivisionToFortranDivisionMapper(IdentityMapper):
    def map_quotient(self, expr):
        return FortranDivision(
                self.rec(expr.numerator),
                self.rec(expr.denominator))


class FTreeWalkerBase:
    def __init__(self, filename):
        from loopy.frontend.fortran.expression import FortranExpressionParser
        self.expr_parser = FortranExpressionParser(self)
        self.filename = filename

    def rec(self, expr, *args, **kwargs):
        mro = list(type(expr).__mro__)
        dispatch_class = kwargs.pop("dispatch_class", type(self))

        while mro:
            method_name = "map_"+mro.pop(0).__name__

            try:
                method = getattr(dispatch_class, method_name)
            except AttributeError:
                pass
            else:
                return method(self, expr, *args, **kwargs)

        raise NotImplementedError(
                "%s does not know how to map type '%s'"
                % (type(self).__name__,
                    type(expr)))

    ENTITY_RE = re.compile(
            r"^(?P<name>[_0-9a-zA-Z]+)\s*"
            r"(\((?P<shape>[-+*/0-9:a-zA-Z, \t]+)\))?"
            r"(\s*=\s*(?P<initializer>.+))?"
            r"$")

    def parse_dimension_specs(self, node, dim_decls):
        def parse_bounds(bounds_str):
            start_end = bounds_str.split(":")

            assert 1 <= len(start_end) <= 2

            return [self.parse_expr(node, s) for s in start_end]

        for decl in dim_decls:
            entity_match = self.ENTITY_RE.match(decl)
            assert entity_match

            groups = entity_match.groupdict()
            name = groups["name"]
            assert name

            if groups["shape"]:
                shape = [parse_bounds(s) for s in groups["shape"].split(",")]
            else:
                shape = None

            init_str = groups["initializer"]
            if init_str:
                init_str = init_str.replace("(/", "[")
                init_str = init_str.replace("/)", "]")
                init_expr = self.parse_expr(node, init_str)

                from numbers import Number
                if isinstance(init_expr, Number):
                    initializer = init_expr
                elif isinstance(init_expr, list):
                    for i, item in enumerate(init_expr):
                        if not isinstance(item, Number):
                            raise LoopyError("unexpected type of "
                                    "item %d in initializer: %s"
                                    % (i+1, type(init_expr).__name__))
                    initializer = init_expr

                else:
                    raise LoopyError("unexpected type of initializer: %s"
                            % type(init_expr).__name__)

            else:
                initializer = None

            yield name, shape, initializer

    def __call__(self, expr, *args, **kwargs):
        return self.rec(expr, *args, **kwargs)

    # {{{ expressions

    def parse_expr(self, node, expr_str, **kwargs):
        try:
            return DivisionToFortranDivisionMapper()(
                    self.expr_parser(expr_str, **kwargs))
        except Exception as e:
            raise LoopyError(
                    "Error parsing expression '%s' on line %d of '%s': %s"
                    % (expr_str, node.item.span[0], self.filename, str(e))) from e

    # }}}

# vim: foldmethod=marker
